@@ -1,5 +1,5 @@
 /* Copyright (C) 2004 MySQL AB
-   Copyright (C) 2004-2017 Alexey Kopytov <akopytov@gmail.com>
+   Copyright (C) 2004-2018 Alexey Kopytov <akopytov@gmail.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -116,10 +116,9 @@ typedef enum
 } file_io_mode_t;
 
 typedef enum {
-  SB_FILE_FLAG_NORMAL,
-  SB_FILE_FLAG_SYNC,
-  SB_FILE_FLAG_DSYNC,
-  SB_FILE_FLAG_DIRECTIO
+  SB_FILE_FLAG_SYNC = 1,
+  SB_FILE_FLAG_DSYNC = 2,
+  SB_FILE_FLAG_DIRECTIO = 4
 } file_flags_t;
 
 #ifdef HAVE_LIBAIO
@@ -163,7 +162,7 @@ static int               file_fsync_end;
 static file_fsync_mode_t file_fsync_mode;
 static float             file_rw_ratio;
 static int               file_merged_requests;
-static long long         file_max_request_size;
+static long long         file_request_size;
 static file_io_mode_t    file_io_mode;
 #ifdef HAVE_LIBAIO
 static unsigned int      file_async_backlog;
@@ -221,8 +220,8 @@ static sb_arg_t fileio_args[] = {
          "number of asynchronous operatons to queue per thread", "128", INT),
 #endif
   SB_OPT("file-extra-flags",
-         "additional flags to use on opening files {sync,dsync,direct}",
-         "", STRING),
+         "list of additional flags to use to open files {sync,dsync,direct}",
+         "", LIST),
   SB_OPT("file-fsync-freq", "do fsync() after this number of requests "
          "(0 - don't use fsync())", "100", INT),
   SB_OPT("file-fsync-all", "do fsync() after each write operation", "off",
@@ -377,7 +376,31 @@ int file_prepare(void)
     {
       log_errno(LOG_FATAL, "Cannot open file '%s'", file_name);
       log_text(LOG_WARNING, "Did you forget to run the prepare step?");
-      return 1; 
+      return 1;
+    }
+
+    if (test_mode == MODE_WRITE)
+      continue;
+
+    /* Validate file size */
+    struct stat buf;
+    if (fstat(files[i], &buf))
+    {
+      log_errno(LOG_FATAL, "fstat() failed on file '%s'", file_name);
+      return 1;
+    }
+    if (buf.st_size < file_size)
+    {
+      char ss1[16], ss2[16];
+      log_text(LOG_FATAL,
+               "Size of file '%s' is %sB, but at least %sB is expected.",
+               file_name,
+               sb_print_value_size(ss1, sizeof(ss1), buf.st_size),
+               sb_print_value_size(ss2, sizeof(ss2), file_size));
+      log_text(LOG_WARNING,
+               "Did you run 'prepare' with different --file-total-size or "
+               "--file-num values?");
+      return 1;
     }
   }
 
@@ -500,22 +523,10 @@ sb_event_t file_get_seq_request(void)
     position= 0;
     current_file= 0;
   }
-  
-  if (file_merged_requests > 0)
-  {
-    if (position + file_max_request_size <= file_size)
-      file_req->size = file_max_request_size;
-    else
-      file_req->size = file_size - position;
-    file_req->file_id = current_file;
-    file_req->pos = position;
-  }
-  else
-  {
-    file_req->size = file_block_size;
-    file_req->file_id = current_file;
-    file_req->pos = position;
-  }
+
+  file_req->file_id = current_file;
+  file_req->pos = position;
+  file_req->size = SB_MIN(file_request_size, file_size - position);
 
   position += file_req->size;
 
@@ -624,7 +635,7 @@ retry:
   tmppos = tmppos - (tmppos % (long long) file_block_size);
   file_req->file_id = (int) (tmppos / (long long) file_size);
   file_req->pos = (long long) (tmppos % (long long) file_size);
-  file_req->size = file_block_size;
+  file_req->size = SB_MIN(file_block_size, file_size - file_req->pos);
 
   if (sb_globals.validate)
   {
@@ -676,7 +687,10 @@ int file_execute_event(sb_event_t *sb_req, int thread_id)
   }
   if (file_req->pos + file_req->size > file_size)
   {
-    log_text(LOG_FATAL, "Too large position discovered in request!");
+    log_text(LOG_FATAL, "I/O request exceeds file size. "
+             "file id: %d file size: %lld req offset: %lld req size: %lld",
+             file_req->file_id, (long long) file_size,
+             (long long) file_req->pos, (long long) file_req->size);
     return 1;
   }
   fd = files[file_req->file_id];
@@ -783,12 +797,21 @@ int file_execute_event(sb_event_t *sb_req, int thread_id)
 
 }
 
+static void print_file_extra_flags(void)
+{
+  log_text(LOG_NOTICE, "Extra file open flags: %s%s%s%s",
+           file_extra_flags == 0 ? "(none)" : "",
+           file_extra_flags & SB_FILE_FLAG_SYNC ? "sync " : "",
+           file_extra_flags & SB_FILE_FLAG_DSYNC ? "dsync ": "",
+           file_extra_flags & SB_FILE_FLAG_DIRECTIO ? "directio" : ""
+           );
+}
 
 void file_print_mode(void)
 {
   char sizestr[16];
-  
-  log_text(LOG_NOTICE, "Extra file open flags: %x", file_extra_flags);
+
+  print_file_extra_flags();
   log_text(LOG_NOTICE, "%d files, %sB each", num_files,
            sb_print_value_size(sizestr, sizeof(sizestr), file_size));
   log_text(LOG_NOTICE, "%sB total file size",
@@ -799,7 +822,7 @@ void file_print_mode(void)
   if (file_merged_requests > 0)
     log_text(LOG_NOTICE, "Merging requests up to %sB for sequential IO.",
              sb_print_value_size(sizestr, sizeof(sizestr),
-                                 file_max_request_size));
+                                 file_request_size));
 
   switch (test_mode)
   {
@@ -959,44 +982,56 @@ const char *get_test_mode_str(file_test_mode_t mode)
 
 static int convert_extra_flags(file_flags_t extra_flags, int *open_flags)
 {
-  switch (extra_flags) {
-  case SB_FILE_FLAG_NORMAL:
+  if (extra_flags == 0)
+  {
 #ifdef _WIN32
     *open_flags = FILE_ATTRIBUTE_NORMAL;
 #endif
-    break;
-  case SB_FILE_FLAG_SYNC:
+  }
+  else
+  {
+    *open_flags = 0;
+
+    if (extra_flags & SB_FILE_FLAG_SYNC)
+    {
 #ifdef _WIN32
-    *open_flags = FILE_FLAG_WRITE_THROUGH;
+      *open_flags |= FILE_FLAG_WRITE_THROUGH;
 #else
-    *open_flags = O_SYNC;
+      *open_flags |= O_SYNC;
 #endif
-    break;
-  case SB_FILE_FLAG_DSYNC:
+    }
+
+    if (extra_flags & SB_FILE_FLAG_DSYNC)
+    {
 #ifdef O_DSYNC
-    *open_flags = O_DSYNC;
+      *open_flags |= O_DSYNC;
 #else
-    log_text(LOG_FATAL,
-             "--file-extra-flags=dsync is not supported on this platform.");
-    return 1;
+      log_text(LOG_FATAL,
+               "--file-extra-flags=dsync is not supported on this platform.");
+      return 1;
 #endif
-    break;
-  case SB_FILE_FLAG_DIRECTIO:
+    }
+
+    if (extra_flags & SB_FILE_FLAG_DIRECTIO)
+    {
 #ifdef HAVE_DIRECTIO
-    /* Will call directio(3) later */
+      /* Will call directio(3) later */
 #elif defined(O_DIRECT)
-    *open_flags = O_DIRECT;
+      *open_flags |= O_DIRECT;
 #elif defined _WIN32
-    *open_flags = FILE_FLAG_NO_BUFFERING;
+      *open_flags |= FILE_FLAG_NO_BUFFERING;
 #else
-    log_text(LOG_FATAL,
-             "--file-extra-flags=direct is not supported on this platform.");
-    return 1;
+      log_text(LOG_FATAL,
+               "--file-extra-flags=direct is not supported on this platform.");
+      return 1;
 #endif
-    break;
-  default:
-    log_text(LOG_FATAL, "Unknown extra flags value: %d", (int) extra_flags);
-    return 1;
+    }
+
+    if (extra_flags > SB_FILE_FLAG_DIRECTIO)
+    {
+      log_text(LOG_FATAL, "Unknown extra flags value: %d", (int) extra_flags);
+      return 1;
+    }
   }
 
   return 0;
@@ -1004,7 +1039,6 @@ static int convert_extra_flags(file_flags_t extra_flags, int *open_flags)
 
 /* Create files of necessary size for test */
 
-
 int create_files(void)
 {
   unsigned int       i;
@@ -1020,7 +1054,7 @@ int create_files(void)
            (long)(file_size / 1024),
            (long)((file_size * num_files) / (1024 * 1024)));
   log_text(LOG_NOTICE, "Creating files for the test...");
-  log_text(LOG_NOTICE, "Extra file open flags: %x", file_extra_flags);
+  print_file_extra_flags();
 
   if (convert_extra_flags(file_extra_flags, &flags))
     return 1;
@@ -1853,25 +1887,30 @@ int parse_arguments(void)
   }
 
   if (file_merged_requests > 0)
-    file_max_request_size = file_block_size * file_merged_requests;
+    file_request_size = file_block_size * file_merged_requests;
   else
-    file_max_request_size = file_block_size;
+    file_request_size = file_block_size;
 
   mode = sb_get_value_string("file-extra-flags");
-  if (mode == NULL || !strlen(mode))
-    file_extra_flags = SB_FILE_FLAG_NORMAL;
-  else if (!strcmp(mode, "sync"))
-    file_extra_flags = SB_FILE_FLAG_SYNC;
-  else if (!strcmp(mode, "dsync"))
-    file_extra_flags = SB_FILE_FLAG_DSYNC;
-  else if (!strcmp(mode, "direct"))
-    file_extra_flags = SB_FILE_FLAG_DIRECTIO;
-  else
+
+  sb_list_item_t *pos;
+  SB_LIST_FOR_EACH(pos, sb_get_value_list("file-extra-flags"))
   {
-    log_text(LOG_FATAL, "Invalid value for file-extra-flags: %s", mode);
-    return 1;
+    const char *val = SB_LIST_ENTRY(pos, value_t, listitem)->data;
+
+    if (!strcmp(val, "sync"))
+      file_extra_flags |= SB_FILE_FLAG_SYNC;
+    else if (!strcmp(val, "dsync"))
+      file_extra_flags |= SB_FILE_FLAG_DSYNC;
+    else if (!strcmp(val, "direct"))
+      file_extra_flags |= SB_FILE_FLAG_DIRECTIO;
+    else
+    {
+      log_text(LOG_FATAL, "Invalid value for file-extra-flags: %s", mode);
+      return 1;
+    }
   }
-  
+
   file_fsync_freq = sb_get_value_int("file-fsync-freq");
   if (file_fsync_freq < 0)
   {
@@ -1916,13 +1955,13 @@ int parse_arguments(void)
   per_thread = malloc(sizeof(*per_thread) * sb_globals.threads);
   for (i = 0; i < sb_globals.threads; i++)
   {
-    per_thread[i].buffer = sb_memalign(file_max_request_size, sb_getpagesize());
+    per_thread[i].buffer = sb_memalign(file_request_size, sb_getpagesize());
     if (per_thread[i].buffer == NULL)
     {
       log_text(LOG_FATAL, "Failed to allocate a memory buffer");
       return 1;
     }
-    memset(per_thread[i].buffer, 0, file_max_request_size);
+    memset(per_thread[i].buffer, 0, file_request_size);
   }
 
   return 0;
@@ -2013,7 +2052,7 @@ static FILE_DESCRIPTOR sb_open(const char *name)
 #endif
 
 #ifdef HAVE_DIRECTIO
-  if (VALID_FILE(file) && file_extra_flags == SB_FILE_FLAG_DIRECTIO &&
+  if (VALID_FILE(file) && file_extra_flags & SB_FILE_FLAG_DIRECTIO &&
       directio(file, DIRECTIO_ON))
   {
     log_errno(LOG_FATAL, "directio() failed");
